// Copyright 2015 Open Source Robotics Foundation, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef FVIZ_CASE_STUDY__FVIZ_HPP_
#define FVIZ_CASE_STUDY__FVIZ_HPP_

#if defined(WIN32)
#ifdef FVIZ_BUILDING_DLL
#define FVIZ_PUBLIC __declspec(dllexport)
#else
#define FVIZ_PUBLIC __declspec(dllimport)
#endif
#else
#define FVIZ_PUBLIC
#endif


#include <string>

FVIZ_PUBLIC
void foo(std::string msg);

#endif  // FVIZ_CASE_STUDY__FVIZ_HPP_
