// Copyright 2015 Open Source Robotics Foundation, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <class_loader/class_loader.hpp>

#include "./fviz.hpp"  // For foo()
#include "./fviz_plugin_base.hpp"

class Bar : public FvizPluginBase
{
public:
  virtual ~Bar() = default;
  void speak()
  {
    foo("from plugin Bar");
  }
};

class Baz : public FvizPluginBase
{
public:
  virtual ~Baz() = default;
  void speak()
  {
    foo("from plugin Baz");
  }
};

CLASS_LOADER_REGISTER_CLASS(Bar, FvizPluginBase)
CLASS_LOADER_REGISTER_CLASS(Baz, FvizPluginBase)
